@@ -3,23 +3,29 @@ use std::sync::atomic::{AtomicPtr, Ordering};
 use std::sync::Arc;
 
 struct QueueHead<T> {
-    element: T,
+    element: Option<T>,
     next: *mut QueueHead<T>,
 }
 
-struct QueueRoot<T> {
-    in_queue: Arc<AtomicPtr<QueueHead<T>>>,
-    out_queue: *mut QueueHead<T>,
-}
-
 pub struct QueueSender<T> {
     in_queue: Arc<AtomicPtr<QueueHead<T>>>,
 }
 
+/* @TODO
+ *
+ * Pushing items into a queue currently requires a load
+ * followed by 1 or more compare_and_swap.
+ *
+ * It should be possible to remove this load and increase
+ * performance in some cases.
+ *
+ * See https://github.com/Amanieu/atomic-rs/blob/master/src/ops.rs
+ */
+
 impl<T> QueueSender<T> {
     pub fn push(&self, element: T) {
         let mut new = Box::into_raw(Box::new(QueueHead {
-            element,
+            element: Some(element),
             next: ptr::null_mut(),
         }));
 
@@ -27,6 +33,11 @@ impl<T> QueueSender<T> {
             unsafe {
                 let in_queue = self.in_queue.load(Ordering::SeqCst);
 
+                if !in_queue.is_null() && (*in_queue).element.is_none() {
+                    drop(Box::from_raw(new));
+                    return;
+                }
+
                 (*new).next = in_queue;
 
                 if self
@@ -34,7 +45,7 @@ impl<T> QueueSender<T> {
                     .compare_and_swap(in_queue, new, Ordering::SeqCst)
                     == in_queue
                 {
-                    break;
+                    return;
                 }
             }
         }
@@ -53,24 +64,22 @@ unsafe impl<T> Sync for QueueSender<T> {}
 
 unsafe impl<T> Send for QueueSender<T> {}
 
-unsafe impl<T> Send for QueueReceiver<T> {}
-
 pub struct QueueReceiver<T> {
-    root: QueueRoot<T>,
+    in_queue: Arc<AtomicPtr<QueueHead<T>>>,
+    out_queue: *mut QueueHead<T>,
 }
 
 impl<T> QueueReceiver<T> {
     pub fn pop(&mut self) -> Option<T> {
-        if self.root.out_queue.is_null() {
+        if self.out_queue.is_null() {
             loop {
-                let mut head = self.root.in_queue.load(Ordering::SeqCst);
+                let mut head = self.in_queue.load(Ordering::SeqCst);
 
                 if head.is_null() {
                     break;
                 }
 
                 if self
-                    .root
                     .in_queue
                     .compare_and_swap(head, ptr::null_mut(), Ordering::SeqCst)
                     == head
@@ -78,8 +87,8 @@ impl<T> QueueReceiver<T> {
                     while !head.is_null() {
                         unsafe {
                             let next = (*head).next;
-                            (*head).next = self.root.out_queue;
-                            self.root.out_queue = head;
+                            (*head).next = self.out_queue;
+                            self.out_queue = head;
                             head = next;
                         }
                     }
@@ -89,18 +98,47 @@ impl<T> QueueReceiver<T> {
             }
         }
 
-        if self.root.out_queue.is_null() {
+        if self.out_queue.is_null() {
             None
         } else {
             unsafe {
-                let head = Box::from_raw(self.root.out_queue);
-                self.root.out_queue = head.next;
-                Some(head.element)
+                let head = Box::from_raw(self.out_queue);
+                self.out_queue = head.next;
+                Some(head.element.unwrap())
             }
         }
     }
 }
 
+impl<T> Drop for QueueReceiver<T> {
+    fn drop(&mut self) {
+        while let Some(_) = self.pop() {}
+
+        let mut ultimate = Box::into_raw(Box::new(QueueHead {
+            element: None,
+            next: ptr::null_mut(),
+        }));
+
+        loop {
+            unsafe {
+                let in_queue = self.in_queue.load(Ordering::SeqCst);
+
+                (*ultimate).next = in_queue;
+
+                if self
+                    .in_queue
+                    .compare_and_swap(in_queue, ultimate, Ordering::SeqCst)
+                    == in_queue
+                {
+                    return;
+                }
+            }
+        }
+    }
+}
+
+unsafe impl<T> Send for QueueReceiver<T> {}
+
 pub struct Queue;
 
 impl Queue {
@@ -113,15 +151,13 @@ impl Queue {
     pub fn unbounded<T>() -> (QueueSender<T>, QueueReceiver<T>) {
         let in_queue = Arc::new(AtomicPtr::new(ptr::null_mut()));
 
-        let root = QueueRoot {
+        let receiver = QueueReceiver {
             in_queue: in_queue.clone(),
             out_queue: ptr::null_mut(),
         };
 
         let sender = QueueSender { in_queue };
 
-        let receiver = QueueReceiver { root };
-
         (sender, receiver)
     }
 }
@@ -129,7 +165,7 @@ impl Queue {
 #[cfg(test)]
 mod tests {
     use crate::Queue;
-    use std::sync::atomic::{AtomicUsize, Ordering};
+    use std::sync::atomic::{AtomicBool, AtomicUsize, Ordering};
     use std::sync::Arc;
     use std::thread;
     use std::time;
@@ -184,6 +220,31 @@ mod tests {
         assert_eq!(rx.pop(), Some(1));
     }
 
+    #[test]
+    fn test_disconnected() {
+        struct MyStruct {
+            dropped: Arc<AtomicBool>,
+        }
+
+        impl Drop for MyStruct {
+            fn drop(&mut self) {
+                self.dropped.store(true, Ordering::SeqCst);
+            }
+        }
+
+        let dropped = Arc::new(AtomicBool::new(false));
+
+        let (tx, rx) = Queue::unbounded();
+
+        drop(rx);
+
+        tx.push(MyStruct {
+            dropped: dropped.clone(),
+        });
+
+        assert!(dropped.load(Ordering::SeqCst));
+    }
+
     #[test]
     fn test_drop() {
         struct MyStruct {
@@ -227,7 +288,7 @@ mod tests {
     fn benchmark() {
         let num_items = 1_000_000;
 
-        for n in [1, 2, 4, 8, 16, 32].iter() {
+        for n in [1, 2, 4, 8, 16, 32, 64, 128].iter() {
             for _ in 0..1 {
                 run(num_items, *n);
             }
